/*
Module : CMMSystem.cpp
Purpose: Defines the implementation for C++ wrapper classes for the Win32 Multimedia APIs
Created: PJN / 27-07-2005
History: PJN / 26-06-2006 1. Make AfxThrow*Exception classes part of the their respective 
                          classes.
                          2. Code now uses new C++ style casts rather than old style C casts where necessary. 
                          3. Updated copyright details
                          4. Inclusion of a CMMSYSTEM_EXT_CLASS preprocessor define to allow the class to be
                          easily used in an extension dll.
                          5. Optimized most of the constructor methods
                          6. Made CWaveIn::GetBuffersFilled thread safe
                          7. Made CWaveOut::GetBuffersFree thread safe
                          8. Made the first parameter to CWaveOut::Write a "const void*".
                          9. Made the first parameter to CWaveIn::Read a "void*".
                          10. Updated code to ensure it clean(ish) compiles on VC 2005.
         PJN / 19-03-2008 1. Updated copyright details
                          2. Updated the sample apps to clean compile on VC 2008
                          3. Remove VC 6 style classwizard style comments from sample apps
                          4. All Sleep calls now use 0 as their parameter to specify that the rest of the
                          time slice should be waited for
         PJN / 24-01-2009 1. Updated copyright details
                          2. The code has now been updated to support VC 2005 or later only. 
                          3. Updated code to compile correctly using _ATL_CSTRING_EXPLICIT_CONSTRUCTORS define
                          4. Updated code to clean compile on VC 2008
                          5. Code now compiles cleanly using Code Analysis (/analyze)
         PJN / 12-03-2016 1. Updated copyright details.
                          2. Update the project settings to more modern defaults.
                          3. Reworked the classes to optionally compile without MFC. By default the class now 
                          use STL classes and idioms but if you define CMMSYSTEM_MFC_EXTENSIONS the class 
                          will revert back to the MFC behaviour.
                          4. Added SAL annotations to all the code.
                          5. Updated the code to clean compile on VC 2010 - 2015
                          6. Updated CWaveOut::GetPosition to allow the "cbmmt" parameter to waveOutGetPosition
                          to be specified.
                          7. Updated CWaveOut::GetDevCaps to allow the "cbwoc" parameter to waveOutGetDevCaps to 
                          be specified.
                          8. Updated CWaveIn::GetPosition to allow the "cbmmt" parameter to waveInGetPosition to 
                          be specified.
                          9. Updated CWaveIn::GetDevCaps to allow the "cbwic" parameter to waveOutGetDevCaps to 
                          be specified.
                          10. Updated CMixer::GetDevCaps to allow the "cbmxcaps" parameter to waveOutGetDevCaps to 
                          be specified.
         PJN / 01-01-2017 1. Updated copyright details.
                          2. Updated the ShowInput example application to include a Direct2D codepath in 
                          addition to the existing GDI codepath. Please note that this optimization requires
                          at least VC 2010 to compile as it makes use of the D2D integration added to MFC. This 
                          optimization helps improve the performance of this application as well as demonstrating 
                          migration of code from GDI to D2D. By default D2D in now used in this application but 
                          if you want to revert back to GDI you can define the CSHOWINPUT_NOD2D preprocessor value.
                          3. Updated the ShowInput example application to include a codepath which uses NVIDIA
                          CUDA to optimize the FFT logic using the cuFFT library. To enable support for this you 
                          will need to have the NVIDIA CUDA Toolkit (https://developer.nvidia.com/cuda-toolkit) 
                          installed. You will also need to define the CSHOWINPUT_USECUDA preprocessor value in 
                          the ShowInput application. A prebuilt x64 CUDA enabled D2D implementation is now 
                          provided in the download. Please note that cuFFT is only available in CUDA when 
                          compiled for x64 as noted here: 
                          http://stackoverflow.com/questions/29889897/cufft-lib-for-win32-is-missing
                          3. Updated the ShowInput application to use a 44.1Khz input bit rate for the audio.
                          At this sampling rate and performing an FFT every 1024 samples, each FFT bar shown in the
                          spectrum corresponds to 42.06 Hertz. Also due to changing to this higher sampling rate 
                          and the 1024 samples used for the FFT, it means that the FFT in this app now 
                          updates at a rate of 43 frames per second instead of the previous 10 frames per second
                          when it was using a sampling rate of 11025Khz.
                          4. Updated the ShowInput application to use wider rectangles to show the frequency
                          spectrum.
                          5. The ShowInput sample application now shows the frequency spectrum normalized. This
                          means that the frequency with the largest amplitude is now shown with a height equal
                          to the client area of the application.
         PJN / 23-12-2017 1. Replaced CString::operator LPC*STR() calls throughout the codebase with 
                          CString::GetString calls
                          2. Replaced NULL throughout the codebase with nullptr. This means that the minimum 
                          requirement for the framework is now VC 2010. 
         PJN / 31-12-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes
                          mean that the code will now only compile on VC 2017 or later.
                          3. Removed code which supported now defunct CMMSYSTEM_MFC_EXTENSIONS define
         PJN / 05-06-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 30-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 16-05-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 2005 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "CMMSystem.h"


//////////////////// Statics / Macros /////////////////////////////////////////

#pragma comment(lib, "winmm.lib") //Automatically link to the Multimedia Library


//////////////////// Implementation ///////////////////////////////////////////

CWaveOutException::CWaveOutException(_In_ MMRESULT Error) noexcept : m_Error{Error}
{
}

_Return_type_success_(return != 0) BOOL CWaveOutException::GetErrorMessage(_Out_z_cap_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext) noexcept
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(lpszError != nullptr);

  //We don't support help context at the moment
  if (pnHelpContext != nullptr)
    *pnHelpContext = 0;

  return (waveOutGetErrorText(m_Error, lpszError, nMaxError) == MMSYSERR_NOERROR);
}


CMMIOException::CMMIOException(_In_ MMRESULT Error) noexcept : m_Error{Error}
{
}

#pragma warning(suppress: 26461)
_Return_type_success_(return != 0) BOOL CMMIOException::GetErrorMessage(_Out_z_cap_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext) noexcept
{
  UNREFERENCED_PARAMETER(lpszError);
  UNREFERENCED_PARAMETER(nMaxError);

  //We don't support help context at the moment
  if (pnHelpContext != nullptr)
    *pnHelpContext = 0;

  //Unfortunately there is no mmioGetErrorText method. Why I'm not sure!
  return FALSE;
}


CWaveInException::CWaveInException(_In_ MMRESULT Error) noexcept : m_Error{Error}
{
}

_Return_type_success_(return != 0) BOOL CWaveInException::GetErrorMessage(_Out_z_cap_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext) noexcept
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(lpszError != nullptr);

  //We don't support help context at the moment
  if (pnHelpContext != nullptr)
    *pnHelpContext = 0;

  return (waveInGetErrorText(m_Error, lpszError, nMaxError) == MMSYSERR_NOERROR);
}

#pragma warning(suppress: 26461)
_Return_type_success_(return != 0) BOOL CMixerException::GetErrorMessage(_Out_z_cap_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext) noexcept
{
  UNREFERENCED_PARAMETER(lpszError);
  UNREFERENCED_PARAMETER(nMaxError);

  //We don't support help context at the moment
  if (pnHelpContext != nullptr)
    *pnHelpContext = 0;

  //Unfortunately there is no mixerGetErrorText method. Why I'm not sure!
  return FALSE;
}

CMixerException::CMixerException(_In_ MMRESULT Error) noexcept : m_Error{Error}
{
}


CMMIOFile::CMMIOFile() noexcept : m_hFile{nullptr}
{
}

CMMIOFile::~CMMIOFile()
{
  Close();
}

void CMMIOFile::Open(_Inout_updates_bytes_opt_(128) LPTSTR szFilename, _In_ DWORD dwOpenFlags)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(!IsOpen()); //Please call Close first!

  MMIOINFO Info{}; //The state of the open file
  m_hFile = mmioOpen(szFilename, &Info, dwOpenFlags);
  if (m_hFile == nullptr)
    ThrowMMIOException(Info.wErrorRet);
}

void CMMIOFile::Close(_In_ UINT wFlags) noexcept
{
  if (m_hFile != nullptr)
  {
    mmioClose(m_hFile, wFlags);
    m_hFile = nullptr;
  }
}

BOOL CMMIOFile::IsOpen() const noexcept
{
  return m_hFile != nullptr;
}

void CMMIOFile::Advance(_In_opt_ LPMMIOINFO lpmmioinfo, _In_ UINT wFlags)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{mmioAdvance(m_hFile, lpmmioinfo, wFlags)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMMIOException(mmRes);
}

void CMMIOFile::Ascend(_In_ LPMMCKINFO lpck)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{mmioAscend(m_hFile, lpck, 0)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMMIOException(mmRes);
}

void CMMIOFile::CreateChunk(_In_ LPMMCKINFO lpck, _In_ UINT wFlags)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{mmioCreateChunk(m_hFile, lpck, wFlags)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMMIOException(mmRes);
}

#pragma warning(suppress: 26461)
void CMMIOFile::Descend(_Inout_ LPMMCKINFO lpck, _In_opt_ LPMMCKINFO lpckParent, _In_ UINT wFlags)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{mmioDescend(m_hFile, lpck, lpckParent, wFlags)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMMIOException(mmRes);
}

void CMMIOFile::Flush(_In_ UINT fuFlush)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{mmioFlush(m_hFile, fuFlush)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMMIOException(mmRes);
}

void CMMIOFile::GetInfo(_Out_ LPMMIOINFO lpmmioinfo)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{mmioGetInfo(m_hFile, lpmmioinfo, 0)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMMIOException(mmRes);
}

LONG CMMIOFile::Read(_Out_writes_bytes_(nCount) void* lpBuf, _In_ LONG nCount)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const LONG nRead{mmioRead(m_hFile, static_cast<char*>(lpBuf), nCount)};
  if (nRead == -1)
    ThrowMMIOException(MMSYSERR_ERROR);

  return nRead;
}

LONG CMMIOFile::Seek(_In_ LONG lOffset, _In_ int iOrigin)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const LONG nPosition{mmioSeek(m_hFile, lOffset, iOrigin)};
  if (nPosition == -1)
    ThrowMMIOException(MMSYSERR_ERROR);

  return nPosition;
}

LRESULT CMMIOFile::SendMessage(_In_ UINT wMsg, _In_ LPARAM lParam1, _In_ LPARAM lParam2) noexcept
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  return mmioSendMessage(m_hFile, wMsg, lParam1, lParam2);
}

void CMMIOFile::SetBuffer(_Out_writes_opt_(cchBuffer) LPSTR pchBuffer, _In_ LONG cchBuffer)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{mmioSetBuffer(m_hFile, pchBuffer, cchBuffer, 0)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMMIOException(mmRes);
}

#pragma warning(suppress: 26461)
void CMMIOFile::SetInfo(_In_ LPMMIOINFO lpmmioinfo)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{mmioSetInfo(m_hFile, lpmmioinfo, 0)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMMIOException(mmRes);
}

LONG CMMIOFile::Write(_In_reads_bytes_(nSize) const void* pbyData, _In_ LONG nSize)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const LONG nWritten{mmioWrite(m_hFile, static_cast<const char*>(pbyData), nSize)};
  if (nWritten == -1)
    ThrowMMIOException(MMSYSERR_ERROR);

  return nWritten;
}

FOURCC CMMIOFile::StringToFOURCC(_In_ LPCTSTR sz, _In_ UINT wFlags) noexcept
{
  return mmioStringToFOURCC(sz, wFlags);
}

void CMMIOFile::Rename(_In_ LPCTSTR szFilename, _In_ LPCTSTR szNewFilename, _In_opt_ LPCMMIOINFO lpmmioinfo)
{
  const MMRESULT mmRes{mmioRename(szFilename, szNewFilename, lpmmioinfo, 0)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMMIOException(mmRes);
}

LPMMIOPROC CMMIOFile::InstallIOproc(_In_ FOURCC fccIOProc, _In_opt_ LPMMIOPROC pIOProc, _In_ DWORD dwFlags) noexcept
{
  return mmioInstallIOProc(fccIOProc, pIOProc, dwFlags);
}

void CMMIOFile::ThrowMMIOException(_In_ MMRESULT Error)
{
  CMMIOException e{Error};
  ATLTRACE(_T("Warning: throwing CMMIOException for error %u\n"), Error);
  throw e;
}


CWaveOut::CWaveOut() noexcept : m_pBlocks{nullptr},
                                m_lBlocks{0},
                                m_dwBlockSize{0},
                                m_lFreeBlocks{0},
                                m_lCurrentBlock{0},
                                m_hWaveOut{nullptr}
{
}

CWaveOut::~CWaveOut()
{
  Close();
  FreeBuffers();
}

void CWaveOut::CreateBuffers(_In_ DWORD dwBlockSize, _In_ LONG lNumBlocks)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(!IsOpen()); //CreateBuffers should be called before you call Open

  //Free up the buffers if already allocated
  if (m_pBlocks != nullptr)
    FreeBuffers();

  //Allocate memory for the entire set in one go from the process heap
  //The layout of this block is all the WAVEHDR's allocated first and then
  //all the buffers for each block follows
#pragma warning(suppress: 26472)
  const auto dwTotalBufferSize{static_cast<DWORD>((sizeof(WAVEHDR) + dwBlockSize) * lNumBlocks)};
  m_pBlocks = static_cast<WAVEHDR*>(HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, dwTotalBufferSize));
  if (m_pBlocks == nullptr)
    ThrowWaveOutException(MMSYSERR_NOMEM);

  //setup the individual WAVERHDR pointers inside the continous blob allocated above
#pragma warning(suppress: 26477)
  ATLASSUME(m_pBlocks != nullptr);
  WAVEHDR* pCurrentHeader{m_pBlocks};
#pragma warning(suppress: 26481 26490)
  LPSTR pCurrentBuffer{reinterpret_cast<LPSTR>(m_pBlocks) + (sizeof(WAVEHDR) * lNumBlocks)};
  for (LONG i{0}; i<lNumBlocks; i++)
  {
#pragma warning(suppress: 26489)
    pCurrentHeader->dwBufferLength = dwBlockSize;
#pragma warning(suppress: 26489)
    pCurrentHeader->lpData = pCurrentBuffer;

    //Prepare for the next loop
#pragma warning(suppress: 26481)
    ++pCurrentHeader;
#pragma warning(suppress: 26481)
    pCurrentBuffer += dwBlockSize;
  }

  //Set the member variables which manage the array
  m_lBlocks = lNumBlocks;
  m_lFreeBlocks = lNumBlocks;
  m_dwBlockSize = dwBlockSize;
  m_lCurrentBlock = 0;
}

void CWaveOut::FreeBuffers() noexcept
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(!IsOpen()); //It's a bad idea to free the buffers while the audio driver
                        //may still be accessing them. You should first call Close
                        //to ensure that the buffers are safe to delete

  if (m_pBlocks != nullptr)
  {
    HeapFree(GetProcessHeap(), 0, m_pBlocks);

    m_pBlocks = nullptr;
    m_lBlocks = 0;
    m_dwBlockSize = 0;
    m_lFreeBlocks = 0;
    m_lCurrentBlock = 0;
  }
}

void CWaveOut::Open(_In_ LPCWAVEFORMATEX pwfx, _In_ UINT uDeviceID)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(!IsOpen()); //Please call Close first!

#pragma warning(suppress: 26490)
  const MMRESULT mmRes{waveOutOpen(&m_hWaveOut, uDeviceID, pwfx, reinterpret_cast<DWORD_PTR>(_CallbackProc), reinterpret_cast<DWORD_PTR>(this), CALLBACK_FUNCTION)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);
}

void CWaveOut::Close() noexcept
{
  if (m_hWaveOut)
  {
#pragma warning(suppress: 26477)
    ATLASSERT(m_lBlocks);
#pragma warning(suppress: 26477)
    ATLASSUME(m_pBlocks != nullptr);

    //wait for all blocks to become used
    while (m_lFreeBlocks < m_lBlocks)
      Sleep(0);

    //unprepare any blocks that are still prepared
    for (LONG i{0}; i<m_lBlocks; i++)
    {
#pragma warning(suppress: 26481)
      if (m_pBlocks[i].dwFlags & WHDR_PREPARED)
#pragma warning(suppress: 26481)
        waveOutUnprepareHeader(m_hWaveOut, &m_pBlocks[i], sizeof(WAVEHDR));
    }

    waveOutClose(m_hWaveOut);
    m_hWaveOut = nullptr;
  }
}

BOOL CWaveOut::IsOpen() const noexcept
{
  return m_hWaveOut != nullptr;
}

void CWaveOut::Write(_In_reads_bytes_(dwSize) const void* pData, _In_ DWORD dwSize)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const BYTE* pbyData{static_cast<const BYTE*>(pData)};

  //The pointer to the current buffer we will be operating on in the loop below
#pragma warning(suppress: 26481)
  WAVEHDR* pCurrentBlock{&m_pBlocks[m_lCurrentBlock]};

  while (dwSize)
  {
#pragma warning(suppress: 26477)
    ATLASSUME(pCurrentBlock != nullptr);

    //first make sure the header we're going to use is unprepared
    if (pCurrentBlock->dwFlags & WHDR_PREPARED)
    {
#pragma warning(suppress: 26486)
      const MMRESULT mmRes{waveOutUnprepareHeader(m_hWaveOut, pCurrentBlock, sizeof(WAVEHDR))};
      if (mmRes != MMSYSERR_NOERROR)
        ThrowWaveOutException(mmRes);
    }

    //If we cannot fully fill up the current block, then simply memcpy
    //the amount presented to us and return (after using the dwUser
    //member variable of the block to indicate how much is in the buffer)
    if (dwSize < (m_dwBlockSize - pCurrentBlock->dwUser)) 
    {
#pragma warning(suppress: 26481 26486 26489)
      memcpy(pCurrentBlock->lpData + pCurrentBlock->dwUser, pbyData, dwSize);
      pCurrentBlock->dwUser += dwSize;
      break;
    }

#pragma warning(suppress: 26472)
    const DWORD dwRemainingSpaceInThisBuffer{m_dwBlockSize - static_cast<DWORD>(pCurrentBlock->dwUser)};
#pragma warning(suppress: 26481 26486 26489)
    memcpy(pCurrentBlock->lpData + pCurrentBlock->dwUser, pbyData, dwRemainingSpaceInThisBuffer);
    dwSize -= dwRemainingSpaceInThisBuffer;
#pragma warning(suppress: 26481)
    pbyData += dwRemainingSpaceInThisBuffer;

    //Prepare the new header
#pragma warning(suppress: 26486)
    MMRESULT mmRes{waveOutPrepareHeader(m_hWaveOut, pCurrentBlock, sizeof(WAVEHDR))};
    if (mmRes != MMSYSERR_NOERROR)
      ThrowWaveOutException(mmRes);

    //And write it out to the driver
    mmRes = waveOutWrite(m_hWaveOut, pCurrentBlock, sizeof(WAVEHDR));
    if (mmRes != MMSYSERR_NOERROR)
      ThrowWaveOutException(mmRes);

    //Decrement the number of free blocks
    --m_lFreeBlocks;

    //wait for a block to become free before we loop around trying to use it
    while (!m_lFreeBlocks)
      Sleep(0);

    //point to the next block, before we loop around
    m_lCurrentBlock = ++m_lCurrentBlock % m_lBlocks;
#pragma warning(suppress: 26481)
    pCurrentBlock = &m_pBlocks[m_lCurrentBlock];
    pCurrentBlock->dwUser = 0;
  }
}

void CWaveOut::GetDevCaps(_Out_ LPWAVEOUTCAPS pwoc, _In_ UINT cbwoc) const
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

#pragma warning(suppress: 26490)
  GetDevCaps(reinterpret_cast<UINT_PTR>(m_hWaveOut), pwoc, cbwoc);
}

void CWaveOut::GetDevCaps(_In_ UINT_PTR uDeviceID, _Out_ LPWAVEOUTCAPS pwoc, _In_ UINT cbwoc)
{
  const MMRESULT mmRes{waveOutGetDevCaps(uDeviceID, pwoc, cbwoc)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);
}

UINT CWaveOut::GetNumDevs() noexcept
{
  //Simply call thro to the waveOutGetNumDevs function, we are not
  //adding any functionality here, but it means that client apps
  //should never have a need to call the waveOut.. API directly
  //by wrapping all the methods.
  return waveOutGetNumDevs();
}

void CWaveOut::BreakLoop()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{waveOutBreakLoop(m_hWaveOut)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);
}

DWORD CWaveOut::GetPitch() const
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  DWORD dwPitch{0};
  const MMRESULT mmRes{waveOutGetPitch(m_hWaveOut, &dwPitch)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);

  return dwPitch;
}

void CWaveOut::SetPitch(_In_ DWORD dwPitch)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{waveOutSetPitch(m_hWaveOut, dwPitch)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);
}

void CWaveOut::SetPlaybackRate(_In_ DWORD dwRate)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{waveOutSetPlaybackRate(m_hWaveOut, dwRate)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);
}

DWORD CWaveOut::GetPlaybackRate() const
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  DWORD dwPlaybackRate{0};
  const MMRESULT mmRes{waveOutGetPlaybackRate(m_hWaveOut, &dwPlaybackRate)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);

  return dwPlaybackRate;
}

void CWaveOut::GetPosition(_Inout_updates_bytes_(cbmmt) LPMMTIME pmmt, _In_ UINT cbmmt) const
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{waveOutGetPosition(m_hWaveOut, pmmt, cbmmt)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);
}

DWORD CWaveOut::GetVolume() const
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  DWORD dwVolume{0};
  const MMRESULT mmRes{waveOutGetVolume(m_hWaveOut, &dwVolume)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);

  return dwVolume;
}

void CWaveOut::SetVolume(_In_ DWORD dwVolume)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{waveOutSetVolume(m_hWaveOut, dwVolume)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);
}

void CWaveOut::Message(_In_ UINT uMsg, _In_ DWORD_PTR dwParam1, _In_ DWORD_PTR dwParam2)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{waveOutMessage(m_hWaveOut, uMsg, dwParam1, dwParam2)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);
}

void CWaveOut::Pause()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{waveOutPause(m_hWaveOut)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);
}

void CWaveOut::Reset()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{waveOutReset(m_hWaveOut)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);
}

void CWaveOut::Restart()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{waveOutRestart(m_hWaveOut)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveOutException(mmRes);
}

LONG CWaveOut::GetBuffersFree() noexcept
{
  return m_lFreeBlocks;
}

void CWaveOut::ThrowWaveOutException(_In_ MMRESULT Error)
{
  CWaveOutException e{Error};
  ATLTRACE(_T("Warning: throwing CWaveOutException for error %u\n"), Error);
  throw e;
}

void CALLBACK CWaveOut::_CallbackProc(_In_ HWAVEOUT /*hWaveOut*/, _In_ UINT uMsg, _In_ DWORD_PTR dwInstance, _In_ DWORD dwParam1, _In_ DWORD dwParam2) noexcept
{
  //Convert from the SDK world to the C++ world
#pragma warning(suppress: 26490)
  CWaveOut* pThis{reinterpret_cast<CWaveOut*>(dwInstance)};
#pragma warning(suppress: 26477)
  ATLASSUME(pThis != nullptr);

  //Call the virtual non-static method which derived classes can override
#pragma warning(suppress: 26486)
  pThis->CallbackProc(uMsg, dwParam1, dwParam2);
}

void CALLBACK CWaveOut::CallbackProc(_In_ UINT uMsg, _In_ DWORD /*dwParam1*/, _In_ DWORD /*dwParam2*/) noexcept
{
  //The only message we are interested in is the WOM_DONE. Derived classes
  //are of course free to do something different. What we do with this message 
  //is decrement (in a thread safe manner, since this callback will occur in 
  //the context of a different thread to which the public methods of this class
  //are called) the number of free blocks
  if (uMsg == WOM_DONE)
  {
    //increment the number of free blocks value
    ++m_lFreeBlocks;
  }
}


CWaveIn::CWaveIn() noexcept : m_pBlocks{nullptr},
                              m_lBlocks{0},
                              m_dwBlockSize{0},
                              m_lFilledBlocks{0},
                              m_lCurrentBlock{0},
                              m_hWaveIn{nullptr}
{
}

CWaveIn::~CWaveIn()
{
  Close();
  FreeBuffers();
}

void CWaveIn::CreateBuffers(_In_ DWORD dwBlockSize, _In_ LONG lNumBlocks)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(!IsOpen()); //CreateBuffers should be called before you call Open

  //Free up the buffers if already allocated
  if (m_pBlocks != nullptr)
    FreeBuffers();

  //Allocate memory for the entire set in one go from the process heap
  //The layout of this block is all the WAVEHDR's allocated first and then
  //all the buffers for each block follows
#pragma warning(suppress: 26472)
  const auto dwTotalBufferSize{static_cast<DWORD>((dwBlockSize + sizeof(WAVEHDR)) * lNumBlocks)};
  m_pBlocks = static_cast<WAVEHDR*>(HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, dwTotalBufferSize));
  if (m_pBlocks == nullptr)
    ThrowWaveInException(MMSYSERR_NOMEM);

  //Setup the individual WAVERHDR pointers inside the continous blob allocated above
#pragma warning(suppress: 26477)
  ATLASSUME(m_pBlocks != nullptr);
  WAVEHDR* pCurrentHeader{m_pBlocks};
#pragma warning(suppress: 26481 26490)
  LPSTR pCurrentBuffer{reinterpret_cast<LPSTR>(m_pBlocks) + (sizeof(WAVEHDR) * lNumBlocks)};
  for (LONG i{0}; i<lNumBlocks; i++)
  {
#pragma warning(suppress: 26489)
    pCurrentHeader->dwBufferLength = dwBlockSize;
#pragma warning(suppress: 26489)
    pCurrentHeader->lpData = pCurrentBuffer;

    //Prepare for the next loop
#pragma warning(suppress: 26481)
    ++pCurrentHeader;
#pragma warning(suppress: 26481)
    pCurrentBuffer += dwBlockSize;
  }

  //Set the member variables which manage the array
  m_lBlocks = lNumBlocks;
  m_lFilledBlocks = 0;
  m_dwBlockSize = dwBlockSize;
  m_lCurrentBlock = 0;
}

void CWaveIn::FreeBuffers() noexcept
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(!IsOpen()); //It's a bad idea to free the buffers while the audio driver
                        //may still be accessing them. You should first call Close
                        //to ensure that the buffers are safe to delete

  if (m_pBlocks != nullptr)
  {
    HeapFree(GetProcessHeap(), 0, m_pBlocks);

    m_pBlocks = nullptr;
    m_lBlocks = 0;
    m_dwBlockSize = 0;
    m_lFilledBlocks = 0;
    m_lCurrentBlock = 0;
  }
}

void CWaveIn::Open(_In_ LPCWAVEFORMATEX pwfx, _In_ UINT uDeviceID)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(!IsOpen()); //Please call Close first!

#pragma warning(suppress: 26490)
  const MMRESULT mmRes{waveInOpen(&m_hWaveIn, uDeviceID, pwfx, reinterpret_cast<DWORD_PTR>(_CallbackProc), reinterpret_cast<DWORD_PTR>(this), CALLBACK_FUNCTION)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveInException(mmRes);
}

void CWaveIn::Close() noexcept
{
  if (m_hWaveIn != nullptr)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_lBlocks);
#pragma warning(suppress: 26477)
    ATLASSUME(m_pBlocks != nullptr);

    //wait for all blocks to be marked as returned
    waveInReset(m_hWaveIn);
    while (m_lFilledBlocks < m_lBlocks)
      Sleep(0);

    //unprepare any blocks that are still prepared
    for (LONG i{0}; i<m_lBlocks; i++) 
    {
#pragma warning(suppress: 26481)
      if (m_pBlocks[i].dwFlags & WHDR_PREPARED)
#pragma warning(suppress: 26481)
        waveInUnprepareHeader(m_hWaveIn, &m_pBlocks[i], sizeof(WAVEHDR));
    }

    waveInClose(m_hWaveIn);
    m_hWaveIn = nullptr;
  }
}

BOOL CWaveIn::IsOpen() const noexcept
{
  return m_hWaveIn != nullptr;
}

void CWaveIn::Read(_Out_writes_bytes_(dwSize) void* pData, _In_ DWORD dwSize)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  BYTE* pbyData{static_cast<BYTE*>(pData)};

  //The pointer to the current buffer we will be operating on in the loop below
#pragma warning(suppress: 26481)
  WAVEHDR* pCurrentBlock{&m_pBlocks[m_lCurrentBlock]};

  while (dwSize)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(pCurrentBlock != nullptr);

    //wait for a block to become filled before we try to use it
    while (!m_lFilledBlocks)
      Sleep(0);

    //first make sure the header we're going to use is unprepared
    if (pCurrentBlock->dwFlags & WHDR_PREPARED)
    {
#pragma warning(suppress: 26486)
      const MMRESULT mmRes{waveInUnprepareHeader(m_hWaveIn, pCurrentBlock, sizeof(WAVEHDR))};
      if (mmRes != MMSYSERR_NOERROR)
        ThrowWaveInException(mmRes);
    }

    //If we cannot fully fill up the current block, then simply memcpy
    //the amount presented to us and return (after using the dwUser
    //member variable of the block to indicate how much is in the buffer)
    if (dwSize < (m_dwBlockSize - pCurrentBlock->dwUser))
    {
#pragma warning(suppress: 26481 26486 26489)
      memcpy(pbyData, pCurrentBlock->lpData + pCurrentBlock->dwUser, dwSize);
      pCurrentBlock->dwUser += dwSize;
      break;
    }

#pragma warning(suppress: 26472)
    DWORD dwRemainingSpaceInThisBuffer{m_dwBlockSize - static_cast<DWORD>(pCurrentBlock->dwUser)};
#pragma warning(suppress: 26481 26486 26489)
    memcpy(pbyData, pCurrentBlock->lpData + pCurrentBlock->dwUser, dwRemainingSpaceInThisBuffer);
    dwSize -= dwRemainingSpaceInThisBuffer;
#pragma warning(suppress: 26481)
    pbyData += dwRemainingSpaceInThisBuffer;

    pCurrentBlock->dwUser = 0;

    //Prepare the new header, so that it is again available to the driver
#pragma warning(suppress: 26486 26486)
    MMRESULT mmRes{waveInPrepareHeader(m_hWaveIn, pCurrentBlock, sizeof(WAVEHDR))};
    if (mmRes != MMSYSERR_NOERROR)
      ThrowWaveInException(mmRes);

    //And make it available to the driver again
    mmRes = waveInAddBuffer(m_hWaveIn, pCurrentBlock, sizeof(WAVEHDR));
    if (mmRes != MMSYSERR_NOERROR)
      ThrowWaveInException(mmRes);

    //Decrement the number of filled blocks
    --m_lFilledBlocks;

     //point to the next block, before we loop around
    m_lCurrentBlock = ++m_lCurrentBlock % m_lBlocks;
#pragma warning(suppress: 26481)
    pCurrentBlock = &m_pBlocks[m_lCurrentBlock];
  }
}

void CWaveIn::GetDevCaps(_Out_writes_bytes_(cbwic) LPWAVEINCAPS pwic, _In_ UINT cbwic) const
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

#pragma warning(suppress: 26490)
  GetDevCaps(reinterpret_cast<UINT_PTR>(m_hWaveIn), pwic, cbwic);
}

void CWaveIn::GetDevCaps(_In_ UINT_PTR uDeviceID, _Out_writes_bytes_(cbwic) LPWAVEINCAPS pwic, _In_ UINT cbwic)
{
  const MMRESULT mmRes{waveInGetDevCaps(uDeviceID, pwic, cbwic)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveInException(mmRes);
}

UINT CWaveIn::GetNumDevs() noexcept
{
  //Simply call thro to the waveInGetNumDevs function, we are not
  //adding any functionality here, but it means that client apps
  //should never have a need to call the waveIn.. API directly
  //by wrapping all the methods.
  return waveInGetNumDevs();
}

void CWaveIn::GetPosition(_Inout_updates_bytes_(cbmmt) LPMMTIME pmmt, _In_ UINT cbmmt) const
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{waveInGetPosition(m_hWaveIn, pmmt, cbmmt)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveInException(mmRes);
}

void CWaveIn::Message(_In_ UINT uMsg, _In_ DWORD_PTR dwParam1, _In_ DWORD_PTR dwParam2)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{waveInMessage(m_hWaveIn, uMsg, dwParam1, dwParam2)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveInException(mmRes);
}

void CWaveIn::Stop()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{waveInStop(m_hWaveIn)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveInException(mmRes);
}

void CWaveIn::Reset()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{waveInReset(m_hWaveIn)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveInException(mmRes);
}

void CWaveIn::Start()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  //Prepare all the buffers
  for (LONG i{0}; i<m_lBlocks; i++)
  {
#pragma warning(suppress: 26477 26481)
    ATLASSERT((m_pBlocks[i].dwFlags & WHDR_PREPARED) == 0);

    //Set the full block as being available
#pragma warning(suppress: 26481)
    m_pBlocks[i].dwUser = 0;

    //Prepare the header
#pragma warning(suppress: 26481)
    MMRESULT mmRes{waveInPrepareHeader(m_hWaveIn, &m_pBlocks[i], sizeof(WAVEHDR))};
    if (mmRes != MMSYSERR_NOERROR)
      ThrowWaveInException(mmRes);

    //Add pass it to the driver for use
#pragma warning(suppress: 26481)
    mmRes = waveInAddBuffer(m_hWaveIn, &m_pBlocks[i], sizeof(WAVEHDR));
    if (mmRes != MMSYSERR_NOERROR)
      ThrowWaveInException(mmRes);
  }

  //Finally start input on the device
  const MMRESULT mmRes{waveInStart(m_hWaveIn)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowWaveInException(mmRes);
}

LONG CWaveIn::GetBuffersFilled() noexcept
{
  return m_lFilledBlocks;
}

void CWaveIn::ThrowWaveInException(_In_ MMRESULT Error)
{
  CWaveInException e{Error};
  ATLTRACE(_T("Warning: throwing CWaveInException for error %u\n"), Error);
  throw e;
}

void CALLBACK CWaveIn::_CallbackProc(_In_ HWAVEIN /*hWaveOut*/, _In_ UINT uMsg, _In_ DWORD_PTR dwInstance, _In_ DWORD dwParam1, _In_ DWORD dwParam2) noexcept
{
  //Convert from the SDK world to the C++ world
#pragma warning(suppress: 26490)
  CWaveIn* pThis{reinterpret_cast<CWaveIn*>(dwInstance)};
#pragma warning(suppress: 26477)
  ATLASSUME(pThis != nullptr);

  //Call the virtual non-static method which derived classes can override
#pragma warning(suppress: 26486)
  pThis->CallbackProc(uMsg, dwParam1, dwParam2);
}

void CALLBACK CWaveIn::CallbackProc(_In_ UINT uMsg, _In_ DWORD /*dwParam1*/, _In_ DWORD /*dwParam2*/) noexcept
{
  //The only message we are interested in is the WIM_DATA. Derived classes
  //are of course free to do something different. What we do with this message 
  //is increment (in a thread safe manner, since this callback will occur in 
  //the context of a different thread to which the public methods of this class
  //are called) the number of filled blocks
  if (uMsg == WIM_DATA)
  {
    //increment the number of filled blocks value
    ++m_lFilledBlocks;
  }
}


CMixer::CMixer() noexcept : m_hMixer{nullptr}
{
}

CMixer::~CMixer()
{
  Close();
}

void CMixer::Open(_In_ UINT uMxID, _In_ DWORD_PTR dwCallback, _In_ DWORD_PTR dwInstance, _In_ DWORD fdwOpen)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(!IsOpen()); //Please call Close first!

  const MMRESULT mmRes{mixerOpen(&m_hMixer, uMxID, dwCallback, dwInstance, fdwOpen)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMixerException(mmRes);
}

void CMixer::Close() noexcept
{
  if (m_hMixer != nullptr)
  {
    mixerClose(m_hMixer);
    m_hMixer = nullptr;
  }
}

BOOL CMixer::IsOpen() const noexcept
{
  return m_hMixer != nullptr;
}

void CMixer::GetDevCaps(_Out_writes_bytes_(cbmxcaps) LPMIXERCAPS pmxcaps, _In_ UINT cbmxcaps) const
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

#pragma warning(suppress: 26490)
  GetDevCaps(reinterpret_cast<UINT_PTR>(m_hMixer), pmxcaps, cbmxcaps);
}

void CMixer::GetDevCaps(_In_ UINT_PTR uMxID, _Out_writes_bytes_(cbmxcaps) LPMIXERCAPS pmxcaps, _In_ UINT cbmxcaps)
{
  const MMRESULT mmRes{mixerGetDevCaps(uMxID, pmxcaps, cbmxcaps)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMixerException(mmRes);
}

UINT CMixer::GetNumDevs() noexcept
{
  //Simply call thro to the mixerGetNumDevs function, we are not
  //adding any functionality here, but it means that client apps
  //should never have a need to call the mixer.. API directly
  //by wrapping all the methods.
  return mixerGetNumDevs();
}

void CMixer::Message(_In_ UINT uMsg, _In_ DWORD_PTR dwParam1, _In_ DWORD_PTR dwParam2)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  const MMRESULT mmRes{mixerMessage(m_hMixer, uMsg, dwParam1, dwParam2)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMixerException(mmRes);
}

void CMixer::SetControlDetails(_In_ LPMIXERCONTROLDETAILS pmxcd, _In_ DWORD fdwDetails)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

#pragma warning(suppress: 26490)
  SetControlDetails(reinterpret_cast<HMIXEROBJ>(m_hMixer), pmxcd, fdwDetails | MIXER_OBJECTF_HMIXER);
}

void CMixer::GetControlDetails(_Inout_ LPMIXERCONTROLDETAILS pmxcd, _In_ DWORD fdwDetails)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

#pragma warning(suppress: 26490)
  GetControlDetails(reinterpret_cast<HMIXEROBJ>(m_hMixer), pmxcd, fdwDetails | MIXER_OBJECTF_HMIXER);
}

void CMixer::GetID(_Out_ UINT* puMxId)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

#pragma warning(suppress: 26490)
  GetID(reinterpret_cast<HMIXEROBJ>(m_hMixer), puMxId, MIXER_OBJECTF_HMIXER);
}

void CMixer::GetLineControls(_Inout_ LPMIXERLINECONTROLS pmxlc, _In_ DWORD fdwControls)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

#pragma warning(suppress: 26490)
  GetLineControls(reinterpret_cast<HMIXEROBJ>(m_hMixer), pmxlc, fdwControls | MIXER_OBJECTF_HMIXER);
}

void CMixer::GetLineInfo(_Inout_ LPMIXERLINE pmxl, _In_ DWORD fdwInfo)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

#pragma warning(suppress: 26490)
  GetLineInfo(reinterpret_cast<HMIXEROBJ>(m_hMixer), pmxl, fdwInfo | MIXER_OBJECTF_HMIXER);
}

void CMixer::SetControlDetails(_In_opt_ HMIXEROBJ hmxobj, _In_ LPMIXERCONTROLDETAILS pmxcd, _In_ DWORD fdwDetails)
{
  const MMRESULT mmRes{mixerSetControlDetails(hmxobj, pmxcd, fdwDetails)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMixerException(mmRes);
}

void CMixer::GetControlDetails(_In_opt_ HMIXEROBJ hmxobj, _Inout_ LPMIXERCONTROLDETAILS pmxcd, _In_ DWORD fdwDetails)
{
  const MMRESULT mmRes{mixerGetControlDetails(hmxobj, pmxcd, fdwDetails)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMixerException(mmRes);
}

void CMixer::GetID(_In_opt_ HMIXEROBJ hmxobj, _Out_ UINT* puMxId, _In_ DWORD fdwId)
{
  const MMRESULT mmRes{mixerGetID(hmxobj, puMxId, fdwId)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMixerException(mmRes);
}

void CMixer::GetLineControls(_In_opt_ HMIXEROBJ hmxobj, _Inout_ LPMIXERLINECONTROLS pmxlc, _In_ DWORD fdwControls)
{
  const MMRESULT mmRes{mixerGetLineControls(hmxobj, pmxlc, fdwControls)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMixerException(mmRes);
}

void CMixer::GetLineInfo(_In_opt_ HMIXEROBJ hmxobj, _Inout_ LPMIXERLINE pmxl, _In_ DWORD fdwInfo)
{
  const MMRESULT mmRes{mixerGetLineInfo(hmxobj, pmxl, fdwInfo)};
  if (mmRes != MMSYSERR_NOERROR)
    ThrowMixerException(mmRes);
}

void CMixer::ThrowMixerException(_In_ MMRESULT Error)
{
  CMixerException e{Error};
  ATLTRACE(_T("Warning: throwing CMixerException for error %u\n"), Error);
  throw e;
}
