/*
Module : CMMSystem.h
Purpose: Defines the interface for C++ wrapper classes for the Win32 Multimedia APIs
Created: PJN / 27-07-2005

Copyright (c) 2005 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once

#ifndef __CMMSYSTEM_H__
#define __CMMSYSTEM_H__

#ifndef _INC_MMSYSTEM
#pragma message("To avoid this message, please put MMSystem.h in your pre compiled header (normally stdafx.h)")
#include <MMSystem.h>
#endif //#ifndef _INC_MMSYSTEM

#ifndef _EXCEPTION_
#pragma message("To avoid this message, you should put exception in your pre compiled header (normally stdafx.h)")
#include <exception>
#endif //#ifndef _EXCEPTION_

#ifndef _ATOMIC_
#pragma message("To avoid this message, you should put atomic in your pre compiled header (normally stdafx.h)")
#include <atomic>
#endif //#ifndef _ATOMIC_

#ifndef CMMSYSTEM_EXT_CLASS
#define CMMSYSTEM_EXT_CLASS
#endif //#ifndef CMMSYSTEM_EXT_CLASS


//////////////////// Classes //////////////////////////////////////////

//The exception classes
class CMMSYSTEM_EXT_CLASS CWaveOutException : public std::exception
{
public:
//Constructors / Destructors
  CWaveOutException(_In_ MMRESULT Error) noexcept;

//Methods
  _Return_type_success_(return != 0) BOOL GetErrorMessage(_Out_z_cap_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext = nullptr) noexcept;

//Data members
  MMRESULT m_Error;
};

class CMMSYSTEM_EXT_CLASS CMMIOException : public std::exception
{
public:
//Constructors / Destructors
  CMMIOException(_In_ MMRESULT Error) noexcept;

//Methods
  _Return_type_success_(return != 0) BOOL GetErrorMessage(_Out_z_cap_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext = nullptr) noexcept;

//Data members
  MMRESULT m_Error;
};

class CMMSYSTEM_EXT_CLASS CWaveInException : public std::exception
{
public:
//Constructors / Destructors
  CWaveInException(_In_ MMRESULT Error) noexcept;

//Methods
  _Return_type_success_(return != 0) BOOL GetErrorMessage(_Out_z_cap_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext = nullptr) noexcept;

//Data members
  MMRESULT m_Error;
};

class CMMSYSTEM_EXT_CLASS CMixerException : public std::exception
{
public:
//Constructors / Destructors
  CMixerException(_In_ MMRESULT Error) noexcept;

//Methods
  _Return_type_success_(return != 0) BOOL GetErrorMessage(_Out_z_cap_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext = nullptr) noexcept;

//Data members
  MMRESULT m_Error;
};

//The wrapper class for the mmIO APIs
class CMMSYSTEM_EXT_CLASS CMMIOFile
{
public:
//Constructors / Destructors
  CMMIOFile() noexcept;
  CMMIOFile(const CMMIOFile&) = delete;
  CMMIOFile(CMMIOFile&&) = delete;
  virtual ~CMMIOFile();

//Methods
  CMMIOFile& operator=(const CMMIOFile&) = delete;
  CMMIOFile& operator=(CMMIOFile&&) = delete;
  void Open(_Inout_updates_bytes_opt_(128) LPTSTR szFilename, _In_ DWORD dwOpenFlags);
  void Close(_In_ UINT wFlags = 0) noexcept;
  _NODISCARD BOOL IsOpen() const noexcept;
  void Advance(_In_opt_ LPMMIOINFO lpmmioinfo, _In_ UINT wFlags);
  void Ascend(_In_ LPMMCKINFO lpck);
  void CreateChunk(_In_ LPMMCKINFO lpck, _In_ UINT wFlags);
  void Descend(_Inout_ LPMMCKINFO lpck, _In_opt_ LPMMCKINFO lpckParent, _In_ UINT wFlags);
  void Flush(_In_ UINT fuFlush);
  void GetInfo(_Out_ LPMMIOINFO lpmmioinfo);
  LONG Read(_Out_writes_bytes_(nCount) void* lpBuf, _In_ LONG nCount);
  LONG Seek(_In_ LONG lOffset, _In_ int iOrigin);
  LRESULT SendMessage(_In_ UINT wMsg, _In_ LPARAM lParam1, _In_ LPARAM lParam2) noexcept;
  void SetBuffer(_Out_writes_opt_(cchBuffer) LPSTR pchBuffer, _In_ LONG cchBuffer);
  void SetInfo(_In_ LPMMIOINFO lpmmioinfo);
  LONG Write(_In_reads_bytes_(nSize) const void* pbyData, _In_ LONG nSize);

//Static methods
  static FOURCC StringToFOURCC(_In_ LPCTSTR sz, _In_ UINT wFlags = 0) noexcept;
  static void Rename(_In_ LPCTSTR szFilename, _In_ LPCTSTR szNewFilename, _In_opt_ LPCMMIOINFO lpmmioinfo);
  static LPMMIOPROC InstallIOproc(_In_ FOURCC fccIOProc, _In_opt_ LPMMIOPROC pIOProc, _In_ DWORD dwFlags) noexcept;
  static void ThrowMMIOException(_In_ MMRESULT Error);

protected:
//Member variables
  HMMIO m_hFile; //The file handle we are encapsulating
};

//The wrapper class for the waveOut APIs
class CMMSYSTEM_EXT_CLASS CWaveOut
{
public:
//Constructors / Destructors
  CWaveOut() noexcept;
  CWaveOut(const CWaveOut&) = delete;
  CWaveOut(CWaveOut&&) = delete;
  virtual ~CWaveOut();

//Methods
  CWaveOut& operator=(const CWaveOut&) = delete;
  CWaveOut& operator=(CWaveOut&&) = delete;
  virtual void CreateBuffers(_In_ DWORD dwBlockSize = 4096, _In_ LONG lNumBlocks = 16);
  void FreeBuffers() noexcept;
  void Open(_In_ LPCWAVEFORMATEX pwfx, _In_ UINT uDeviceID = WAVE_MAPPER);
  void Close() noexcept;
  _NODISCARD BOOL IsOpen() const noexcept;
  void Write(_In_reads_bytes_(dwSize) const void* pData, _In_ DWORD dwSize);
  _NODISCARD DWORD GetPitch() const;
  void SetPitch(_In_ DWORD dwPitch);
  void SetPlaybackRate(_In_ DWORD dwRate);
  _NODISCARD DWORD GetPlaybackRate() const;
  _NODISCARD DWORD GetVolume() const;
  void SetVolume(_In_ DWORD dwVolume);
  void GetPosition(_Inout_updates_bytes_(cbmmt) LPMMTIME pmmt, _In_ UINT cbmmt) const;
  void GetDevCaps(_Out_ LPWAVEOUTCAPS pwoc, _In_ UINT cbwoc = sizeof(WAVEOUTCAPS)) const;
  void BreakLoop();
  void Message(_In_ UINT uMsg, _In_ DWORD_PTR dwParam1, _In_ DWORD_PTR dwParam2);
  void Pause();
  void Reset();
  void Restart();
  _NODISCARD LONG GetBuffersFree() noexcept;

//Static methods
  static UINT GetNumDevs() noexcept;
  static void GetDevCaps(_In_ UINT_PTR uDeviceID, _Out_ LPWAVEOUTCAPS pwoc, _In_ UINT cbwoc = sizeof(WAVEOUTCAPS));
  static void ThrowWaveOutException(_In_ MMRESULT Error);

protected:
//Static methods
  static void CALLBACK _CallbackProc(_In_ HWAVEOUT hWaveOut, _In_ UINT uMsg, _In_ DWORD_PTR dwInstance, _In_ DWORD dwParam1, _In_ DWORD dwParam2) noexcept;

//Virtual methods
  virtual void CALLBACK CallbackProc(_In_ UINT uMsg, _In_ DWORD dwParam1, _In_ DWORD dwParam2) noexcept;

//Member variables
  WAVEHDR* m_pBlocks;              //The pointer to the array of blocks we use to manage output
  LONG m_lBlocks;                  //The number of blocks in the m_pBlocks array
  DWORD m_dwBlockSize;             //The size in bytes of each block in the "m_pBlocks" array
  std::atomic<LONG> m_lFreeBlocks; //The current number of free blocks in the "m_pBlocks" array
  LONG m_lCurrentBlock;            //The current block which the code will be using for writing to
  HWAVEOUT m_hWaveOut;             //The handle to the wave output device
};

//The wrapper class for the waveIn APIs
class CMMSYSTEM_EXT_CLASS CWaveIn
{
public:
//Constructors / Destructors
  CWaveIn() noexcept;
  CWaveIn(const CWaveIn&) = delete;
  CWaveIn(CWaveIn&&) = delete;
  virtual ~CWaveIn();

//Methods
  CWaveIn& operator=(const CWaveIn&) = delete;
  CWaveIn& operator=(CWaveIn&&) = delete;
  virtual void CreateBuffers(_In_ DWORD dwBlockSize = 4096, _In_ LONG lNumBlocks = 16);
  void FreeBuffers() noexcept;
  void Open(_In_ LPCWAVEFORMATEX pwfx, _In_ UINT uDeviceID = WAVE_MAPPER);
  void Close() noexcept;
  BOOL IsOpen() const noexcept;
  void Read(_Out_writes_bytes_(dwSize) void* pbyData, _In_ DWORD dwSize);
  void GetPosition(_Inout_updates_bytes_(cbmmt) LPMMTIME pmmt, _In_ UINT cbmmt) const;
  void GetDevCaps(_Out_writes_bytes_(cbwic) LPWAVEINCAPS pwoc, _In_ UINT cbwic = sizeof(WAVEINCAPS)) const;
  void Message(_In_ UINT uMsg, _In_ DWORD_PTR dwParam1, _In_ DWORD_PTR dwParam2);
  void Stop();
  void Reset();
  void Start();
  _NODISCARD LONG GetBuffersFilled() noexcept;

//Static methods
  static UINT GetNumDevs() noexcept;
  static void GetDevCaps(_In_ UINT_PTR uDeviceID, _Out_writes_bytes_(cbwic) LPWAVEINCAPS pwic, _In_ UINT cbwic = sizeof(WAVEINCAPS));
  static void ThrowWaveInException(_In_ MMRESULT Error);

protected:
//Static methods
  static void CALLBACK _CallbackProc(_In_ HWAVEIN hWaveOut, _In_ UINT uMsg, _In_ DWORD_PTR dwInstance, _In_ DWORD dwParam1, _In_ DWORD dwParam2) noexcept;

//Virtual methods
  virtual void CALLBACK CallbackProc(_In_ UINT uMsg, _In_ DWORD dwParam1, _In_ DWORD dwParam2) noexcept;

//Member variables
  WAVEHDR* m_pBlocks;                //The pointer to the array of blocks we use to manage output
  LONG m_lBlocks;                    //The number of blocks in the m_pBlocks array
  DWORD m_dwBlockSize;               //The size in bytes of each block in the "m_pBlocks" array
  std::atomic<LONG> m_lFilledBlocks; //The current number of filled blocks in the "m_pBlocks" array
  LONG m_lCurrentBlock;              //The current block which the code will be using for writing to
  HWAVEIN m_hWaveIn;                 //The handle to the wave in device
};

//The wrapper class for the mixer APIs
class CMMSYSTEM_EXT_CLASS CMixer
{
public:
//Constructors / Destructors
  CMixer() noexcept;
  CMixer(const CMixer&) = delete;
  CMixer(CMixer&&) = delete;
  virtual ~CMixer();

//Methods
  CMixer& operator=(const CMixer&) = delete;
  CMixer& operator=(CMixer&&) = delete;
  void Open(_In_ UINT uMxID, _In_ DWORD_PTR dwCallback = NULL, _In_ DWORD_PTR dwInstance = NULL, _In_ DWORD fdwOpen = MIXER_OBJECTF_MIXER);
  void Close() noexcept;
  _NODISCARD BOOL IsOpen() const noexcept;
  void GetDevCaps(_Out_writes_bytes_(cbmxcaps) LPMIXERCAPS pmxcaps, _In_ UINT cbmxcaps = sizeof(MIXERCAPS)) const;
  void Message(_In_ UINT uMsg, _In_ DWORD_PTR dwParam1, _In_ DWORD_PTR dwParam2);
  void SetControlDetails(_In_ LPMIXERCONTROLDETAILS pmxcd, _In_ DWORD fdwDetails);
  void GetControlDetails(_Inout_ LPMIXERCONTROLDETAILS pmxcd, _In_ DWORD fdwDetails);
  void GetID(_Out_ UINT* puMxId);
  void GetLineControls(_Inout_ LPMIXERLINECONTROLS pmxlc, _In_ DWORD fdwControls); 
  void GetLineInfo(_Inout_ LPMIXERLINE pmxl, _In_ DWORD fdwInfo);

//Static methods
  static void SetControlDetails(_In_opt_ HMIXEROBJ hmxobj, _In_ LPMIXERCONTROLDETAILS pmxcd, _In_ DWORD fdwDetails);
  static void GetControlDetails(_In_opt_ HMIXEROBJ hmxobj, _Inout_ LPMIXERCONTROLDETAILS pmxcd, _In_ DWORD fdwDetails);
  static void GetID(_In_opt_ HMIXEROBJ hmxobj, _Out_ UINT* puMxId, _In_ DWORD fdwId);
  static void GetLineControls(_In_opt_ HMIXEROBJ hmxobj, _Inout_ LPMIXERLINECONTROLS pmxlc, _In_ DWORD fdwControls); 
  static void GetLineInfo(_In_opt_ HMIXEROBJ hmxobj, _Inout_ LPMIXERLINE pmxl, _In_ DWORD fdwInfo);

//Static methods
  static UINT GetNumDevs() noexcept;
  static void GetDevCaps(_In_ UINT_PTR uDeviceID, _Out_writes_bytes_(cbmxcaps) LPMIXERCAPS pmxcaps, _In_ UINT cbmxcaps = sizeof(MIXERCAPS));
  static void ThrowMixerException(_In_ MMRESULT Error);

protected:
//Member variables
  HMIXER m_hMixer; //The handle to the mixer device
};

#endif //#ifndef __CMMSYSTEM_H__
